@@ -4,12 +4,17 @@ from .jusbr_pomes import (
 from .provider_pomes import (
     provider_register, provider_get_token
 )
+from .token_pomes import (
+    token_validate
+)
 
 __all__ = [
     # jusbr_pomes
     "jusbr_setup", "jusbr_get_token", "jusbr_set_scope",
     # provider_pomes
-    "provider_register", "provider_get_token"
+    "provider_register", "provider_get_token",
+    # token_pomes
+    "token_validate"
 ]
 
 from importlib.metadata import version
