@@ -13,55 +13,44 @@ from typing import Any, Final
 
 JUSBR_CLIENT_ID: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_CLIENT_ID")
 JUSBR_CLIENT_SECRET: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_CLIENT_SECRET")
-JUSBR_LOGIN_TIMEOUT: Final[int] = env_get_int(key=f"{APP_PREFIX}_JUSBR_LOGIN_TIMEOUT")
+JUSBR_CLIENT_TIMEOUT: Final[int] = env_get_int(key=f"{APP_PREFIX}_JUSBR_CLIENT_TIMEOUT")
 
-JUSBR_CALLBACK_ENDPOINT: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_CALLBACK_ENDPOINT",
+JUSBR_ENDPOINT_CALLBACK: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_ENDPOINT_CALLBACK",
                                                   def_value="/iam/jusbr:callback")
-JUSBR_TOKEN_ENDPOINT: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_TOKEN_ENDPOINT",
-                                               def_value="/iam/jusbr:get-token")
-JUSBR_LOGIN_ENDPOINT: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_LOGIN_ENDPOINT",
+JUSBR_ENDPOINT_LOGIN: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_ENDPOINT_LOGIN",
                                                def_value="/iam/jusbr:login")
-JUSBR_LOGOUT_ENDPOINT: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_LOGOUT_ENDPOINT",
+JUSBR_ENDPOINT_LOGOUT: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_ENDPOINT_LOGOUT",
                                                 def_value="/iam/jusbr:logout")
+JUSBR_ENDPOINT_TOKEN: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_ENDPOINT_TOKEN",
+                                               def_value="/iam/jusbr:get-token")
 
-JUSBR_CALLBACK_URL: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_CALLBACK_URL")
-JUSBR_AUTH_URL: Final[str] = env_get_str(
-    key=f"{APP_PREFIX}JUSBR_AUTH_URL",
-    def_value="https://sso.stg.cloud.pje.jus.br/auth/realms/pje/protocol/openid-connect/auth"
-)
-JUSBR_TOKEN_URL: Final[str] = env_get_str(
-    key=f"{APP_PREFIX}JUSBR_TOKEN_URL",
-    def_value="https://sso.stg.cloud.pje.jus.br/auth/realms/pje/protocol/openid-connect/token"
-)
+JUSBR_PUBLIC_KEY_LIFETIME: Final[int] = env_get_int(key=f"{APP_PREFIX}_JUSBR_PUBLIC_KEY_LIFETIME",
+                                                    def_value=86400)  # 24 hours
+JUSBR_URL_AUTH_BASE: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_URL_AUTH_BASE")
+JUSBR_URL_AUTH_CALLBACK: Final[str] = env_get_str(key=f"{APP_PREFIX}_JUSBR_URL_AUTH_CALLBACK")
 
-# safe memory cache - structure:
+# registry structure:
 # {
 #    "client-id": <str>,
 #    "client-secret": <str>,
+#    "client-timeout": <int>,
+#    "public_key": <str>,
+#    "key-expiration": <int>,
 #    "auth-url": <str>,
-#    "token-url": <str>,
-#    "login-timeout": <int>,
-#    "users": [
+#    "callback-url": <str>,
+#    "users": {
 #       "<user-id>": {
 #         "cache-obj": <Cache>,
 #         "oauth-scope": <str>,
 #         "access-expiration": <timestamp>,
-#         data in <TTLCache>:
+#         data in <Cache>:
 #           "oauth-state": <str>
 #           "access-token": <str>
 #           "refresh-token": <str>
 #       }
-#    ]
+#    }
 # }
-_jusbr_registry: dict[str, Any] = {
-    "client-id": None,
-    "client-secret": None,
-    "login-timeout": None,
-    "auth-url": None,
-    "callback-url": None,
-    "token-url": None,
-    "users": []
-}
+_jusbr_registry: dict[str, Any] | None = None
 
 # dafault logger
 _logger: Logger | None = None
@@ -70,14 +59,14 @@ _logger: Logger | None = None
 def jusbr_setup(flask_app: Flask,
                 client_id: str = JUSBR_CLIENT_ID,
                 client_secret: str = JUSBR_CLIENT_SECRET,
-                login_timeout: int = JUSBR_LOGIN_TIMEOUT,
-                callback_endpoint: str = JUSBR_CALLBACK_ENDPOINT,
-                token_endpoint: str = JUSBR_TOKEN_ENDPOINT,
-                login_endpoint: str = JUSBR_LOGIN_ENDPOINT,
-                logout_endpoint: str = JUSBR_LOGOUT_ENDPOINT,
-                auth_url: str = JUSBR_AUTH_URL,
-                callback_url: str = JUSBR_CALLBACK_URL,
-                token_url: str = JUSBR_TOKEN_URL,
+                client_timeout: int = JUSBR_CLIENT_TIMEOUT,
+                public_key_lifetime: int = JUSBR_PUBLIC_KEY_LIFETIME,
+                callback_endpoint: str = JUSBR_ENDPOINT_CALLBACK,
+                token_endpoint: str = JUSBR_ENDPOINT_TOKEN,
+                login_endpoint: str = JUSBR_ENDPOINT_LOGIN,
+                logout_endpoint: str = JUSBR_ENDPOINT_LOGOUT,
+                auth_url: str = JUSBR_URL_AUTH_BASE,
+                callback_url: str = JUSBR_URL_AUTH_CALLBACK,
                 logger: Logger = None) -> None:
     """
     Configure the JusBR IAM.
@@ -87,14 +76,14 @@ def jusbr_setup(flask_app: Flask,
     :param flask_app: the Flask application
     :param client_id: the client's identification with JusBR
     :param client_secret: the client's password with JusBR
-    :param login_timeout: timeout for login authentication (in seconds,defaults to no timeout)
+    :param client_timeout: timeout for login authentication (in seconds,defaults to no timeout)
+    :param public_key_lifetime: how long to use JusBR's public key, before refreshing it (in seconds)
     :param callback_endpoint: endpoint for the callback from JusBR
     :param token_endpoint: endpoint for retrieving the JusBR authentication token
     :param login_endpoint: endpoint for redirecting user to JusBR login page
     :param logout_endpoint: endpoint for terminating user access to JusBR
-    :param auth_url: URL to access the JusBR login page
+    :param auth_url: base URL to request the JusBR services
     :param callback_url: URL for JusBR to callback on login
-    :param token_url: URL for obtaing or refreshing the token
     :param logger: optional logger
     """
     # establish the logger
@@ -102,15 +91,17 @@ def jusbr_setup(flask_app: Flask,
     _logger = logger
 
     # configure the JusBR registry
-    global _jusbr_registry  # noqa: PLW0602
-    _jusbr_registry.update({
+    global _jusbr_registry
+    _jusbr_registry = {
         "client-id": client_id,
         "client-secret": client_secret,
-        "login-timeout": login_timeout,
+        "client-timeout": client_timeout,
         "auth-url": auth_url,
         "callback-url": callback_url,
-        "token-url": token_url
-    })
+        "key-expiration": int(datetime.now(tz=TZ_LOCAL).timestamp()),
+        "key-lifetime": public_key_lifetime,
+        "users": {}
+    }
 
     # establish the endpoints
     if token_endpoint:
@@ -132,7 +123,7 @@ def jusbr_setup(flask_app: Flask,
         flask_app.add_url_rule(rule=callback_endpoint,
                                endpoint="jusbr-callback",
                                view_func=service_callback,
-                               methods=["POST"])
+                               methods=["GET", "POST"])
 
 
 # @flask_app.route(rule=<login_endpoint>,  # JUSBR_LOGIN_ENDPOINT: /iam/jusbr:login
@@ -141,21 +132,20 @@ def service_login() -> Response:
     """
     Entry point for the JusBR login service.
 
-    Redirect the request to the JusBR authentication page, with the apprpriate parameters.
+    Redirect the request to the JusBR authentication page, with the appropriate parameters.
 
     :return: the response from the redirect operation
     """
     global _jusbr_registry
 
-    # retrieve user id
+    # retrieve user data (if not provided, 'user_id' is temporarily set to 'oauth_state'
     input_params: dict[str, Any] = request.values
-    user_id: str = input_params.get("user-id") or input_params.get("login")
-
-    # retrieve user data
+    oauth_state: str = "".join(secrets.choice(string.ascii_letters + string.digits) for _ in range(16))
+    user_id: str = input_params.get("user-id") or input_params.get("login") or oauth_state
+    # obtain user data
     user_data: dict[str, Any] = __get_user_data(user_id=user_id,
                                                 logger=_logger)
     # build redirect url
-    oauth_state: str = "".join(secrets.choice(string.ascii_letters + string.digits) for _ in range(16))
     timeout: int = __get_login_timeout()
     safe_cache: Cache
     if timeout:
@@ -165,9 +155,9 @@ def service_login() -> Response:
         safe_cache = FIFOCache(maxsize=16)
     safe_cache["oauth-state"] = oauth_state
     user_data["cache-obj"] = safe_cache
-    auth_url: str = (f"{_jusbr_registry["auth-url"]}?response_type=code"
+    auth_url: str = (f"{_jusbr_registry["auth-url"]}/protocol/openid-connect/auth?response_type=code"
                      f"&client_id={_jusbr_registry["client-id"]}"
-                     f"&redirect_url={_jusbr_registry["callback-url"]}"
+                     f"&redirect_uri={_jusbr_registry["callback-url"]}"
                      f"&state={oauth_state}")
     if user_data.get("oauth-scope"):
         auth_url += f"&scope={user_data.get("oauth-scope")}"
@@ -184,17 +174,17 @@ def service_logout() -> Response:
 
     Remove all data associating the user with JusBR from the registry.
 
-    :return: the response from the redirect operation
+    :return: response *OK*
     """
     global _jusbr_registry
 
     # retrieve user id
-    input_params: dict[str, Any] = request.values
+    input_params: dict[str, Any] = request.args
     user_id: str = input_params.get("user-id") or input_params.get("login")
 
     # remove user data
-    if user_id in _jusbr_registry.get("users"):
-        _jusbr_registry.pop(user_id)
+    if user_id and user_id in _jusbr_registry.get("users"):
+        _jusbr_registry["users"].pop(user_id)
         if _logger:
             _logger.debug(f"User '{user_id}' removed from the registry")
 
@@ -202,7 +192,7 @@ def service_logout() -> Response:
 
 
 # @flask_app.route(rule=<callback_endpoint>,  # JUSBR_CALLBACK_ENDPOINT: /iam/jusbr:callback
-#                  methods=["POST"])
+#                  methods=["GET", "POST"])
 def service_callback() -> Response:
     """
     Entry point for the callback from JusBR on authentication operation.
@@ -210,14 +200,18 @@ def service_callback() -> Response:
     :return: the response containing the token, or *NOT AUTHORIZED*
     """
     global _jusbr_registry
+    from .token_pomes import token_validate
 
     # validate the OAuth2 state
     oauth_state: str = request.args.get("state")
+    user_id: str | None = None
     user_data: dict[str, Any] | None = None
     if oauth_state:
-        for data in _jusbr_registry.get("users"):
-            safe_cache: Cache = user_data.get("cache-obj")
-            if safe_cache and oauth_state == safe_cache.get("oauth-state"):
+        for user, data in _jusbr_registry.get("users").items():
+            safe_cache: Cache = data.get("cache-obj")
+            if user == oauth_state or \
+                    (safe_cache and oauth_state == safe_cache.get("oauth-state")):
+                user_id = user
                 user_data = data
                 # 'oauth-state' is to be used only once
                 safe_cache["oauth-state"] = None
@@ -237,6 +231,20 @@ def service_callback() -> Response:
                              body_data=body_data,
                              errors=errors,
                              logger=_logger)
+        # retrieve the token's claims
+        if not errors:
+            token_claims: dict[str, dict[str, Any]] = token_validate(token=token,
+                                                                     issuer=_jusbr_registry.get("auth-url"),
+                                                                     public_key=_jusbr_registry.get("public_key"),
+                                                                     errors=errors,
+                                                                     logger=_logger)
+            if not errors:
+                token_user: str = token_claims["payload"].get("preferred_username")
+                if user_id == oauth_state:
+                    user_id = token_user
+                    _jusbr_registry["users"][user_id] = _jusbr_registry["users"].pop(oauth_state)
+                elif token_user != user_id:
+                    errors.append(f"Token was issued to user '{token_user}'")
     else:
         msg: str = "Unknown OAuth2 code received"
         if __get_login_timeout():
@@ -248,7 +256,9 @@ def service_callback() -> Response:
         result = jsonify({"errors": "; ".join(errors)})
         result.status_code = 400
     else:
-        result = jsonify({"access_token": token})
+        result = jsonify({
+            "user_id": user_id,
+            "access_token": token})
 
     return result
 
@@ -259,20 +269,22 @@ def service_token() -> Response:
     """
     Entry point for retrieving the JusBR token.
 
-    :return: the response containing the token, or *NOT AUTHORIZED*
+    :return: the response containing the token, or *UNAUTHORIZED*
     """
     # retrieve user id
-    input_params: dict[str, Any] = request.values
+    input_params: dict[str, Any] = request.args
     user_id: str = input_params.get("user-id") or input_params.get("login")
 
     # retrieve the token
+    errors: list[str] = []
     token: str = jusbr_get_token(user_id=user_id,
                                  logger=_logger)
     result: Response
     if token:
         result = jsonify({"token": token})
     else:
-        result = Response(status=401)
+        result = Response("; ".join(errors))
+        result.status_code = 401
 
     return result
 
@@ -293,9 +305,10 @@ def jusbr_get_token(user_id: str,
     # initialize the return variable
     result: str | None = None
 
-    user_data: dict[str, Any] = _jusbr_registry["users"].get(user_id)
-    safe_cache: Cache = user_data["cache-obj"] if user_data else None
-    if user_data and safe_cache:
+    user_data: dict[str, Any] = __get_user_data(user_id=user_id,
+                                                logger=logger)
+    safe_cache: Cache = user_data.get("cache-obj")
+    if safe_cache:
         access_expiration: int = user_data.get("access-expiration")
         now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
         if now < access_expiration:
@@ -345,6 +358,42 @@ def jusbr_set_scope(user_id: str,
         logger.debug(f"Scope for user '{user_id}' set to '{scope}'")
 
 
+def __get_public_key(url: str,
+                     logger: Logger | None) -> str:
+    """
+    Obtain the public key used by JusBR to sign the authentication tokens.
+
+    :param url: the base URL to request the public key
+    :return: the public key, in *PEM* format
+    """
+    from pypomes_crypto import crypto_jwk_convert
+    global _jusbr_registry
+
+    # initialize the return variable
+    result: str | None = None
+
+    now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
+    if now > _jusbr_registry.get("key-expiration"):
+        # obtain a new public key
+        url: str = f"{url}/protocol/openid-connect/certs"
+        response: requests.Response = requests.get(url=url)
+        if response.status_code == 200:
+            # request succeeded
+            reply: dict[str, Any] = response.json()
+            result = crypto_jwk_convert(jwk=reply["keys"][0],
+                                        fmt="PEM")
+            _jusbr_registry["public-key"] = result
+            duration: int = _jusbr_registry.get("key-lifetime") or 0
+            _jusbr_registry["key-expiration"] = now + duration
+        elif logger:
+            logger.error(msg=f"GET '{url}': failed, "
+                             f"status {response.status_code}, reason '{response.reason}'")
+    else:
+        result = _jusbr_registry.get("public-key")
+
+    return result
+
+
 def __get_login_timeout() -> int | None:
     """
     Retrieve the timeout currently applicable for the login operation.
@@ -353,7 +402,7 @@ def __get_login_timeout() -> int | None:
     """
     global _jusbr_registry
 
-    timeout: int = _jusbr_registry.get("login-timeout")
+    timeout: int = _jusbr_registry.get("client-timeout")
     return timeout if isinstance(timeout, int) and timeout > 0 else None
 
 
@@ -390,7 +439,7 @@ def __post_jusbr(user_data: dict[str, Any],
     For code for token exchange, *body_data* will have the attributes
         - "grant_type": "authorization_code"
         - "code": <16-character-random-code>
-        - "redirect_url": <callback-url>
+        - "redirect_uri": <callback-url>
     For token refresh, *body_data* will have the attributes
         - "grant_type": "refresh_token"
         - "refresh_token": <current-refresh-token>
@@ -415,9 +464,10 @@ def __post_jusbr(user_data: dict[str, Any],
     if client_secret:
         body_data["client_secret"] = client_secret
 
+    # obtain the token
     err_msg: str | None = None
     safe_cache: Cache = user_data.get("cache-obj")
-    url: str = _jusbr_registry.get("auth-url")
+    url: str = _jusbr_registry.get("auth-url") + "/protocol/openid-connect/token"
     now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
     try:
         # JusBR return on a token request:
@@ -445,6 +495,8 @@ def __post_jusbr(user_data: dict[str, Any],
             # request resulted in error
             err_msg = (f"POST '{url}': failed, "
                        f"status {response.status_code}, reason '{response.reason}'")
+            if hasattr(response, "content") and response.content:
+                err_msg += f", content '{response.content}'"
             if response.status_code == 401 and "refresh_token" in body_data:
                 # refresh token is no longer valid
                 safe_cache["refresh-token"] = None
